import typing
from threading import Lock, Thread, Condition

import websocket
import json
from functools import reduce


class RpcClient:
    """
    RpcClient represents an WS RPC client, tailored to the
    QIX Associative Engine's RPC API specifically.

    It is safe to use the client.send from multiple threads.
    """

    def __init__(
        self,
        url: str,
        header: typing.List[str] = None,
        request_interceptors=None,
        response_interceptors=None,
    ):
        self._socket = None
        self._id = -1
        self._connected = False
        self._received = None

        if header is None:
            header = []
        if response_interceptors is None:
            response_interceptors = []
        self._response_interceptors = response_interceptors
        if request_interceptors is None:
            request_interceptors = []
        self._request_interceptors = request_interceptors
        self.lock = Lock()
        if not url:
            raise Exception("Empty url")
        self.connect(url, header)

    def _watch_recv(self):
        """
        _watch_recv watches for socket responses.
        Adds the response to _received.
        """

        while True:
            if not self._connected:
                return
            res = self._socket.recv()
            with self._received_added:
                if res:
                    res = json.loads(res)
                    # add response to _received and notify waiting
                    self._received[res["id"]] = res
                    self._received_added.notify_all()

    def connect(self, url: str, header: typing.List[str] = None):
        """
        connect establishes a connection to provided url
        using the specified headers.

        If the client is already connected an exception will
        be raised.
        """
        if header is None:
            header = []

        if self._connected:
            raise Exception("Client already connected")
        socket = websocket.WebSocket()
        socket.connect(url, header=header, suppress_origin=True)
        socket.recv()

        self._socket = socket
        self._received = {}
        self._connected = True
        self._id = -1
        self._received_added = Condition()

        self._watch_recv_thread = Thread(target=self._watch_recv)
        self._watch_recv_thread.start()

    def is_connected(self):
        """
        return connected state
        """

        return self._connected

    def close(self):
        """
        close closes the socket (if it's open).
        """

        if self._connected:
            self._socket.send_close()
            self._connected = False
        if self._watch_recv_thread.is_alive():
            self._watch_recv_thread.join()

    def __enter__(self):
        """
        __enter__ is called when client is used in a 'with' statement.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        __exit__ is called when the 'with' scope is exited. This will call
        the client's close method.
        """

        self.close()

    def send(self, method: str, handle: int, *params):
        """
        send is a thread-safe method that sends a websocket-message with the
        specified method, handle and parameters.
        The resulting response is returned.

        If the client isn't connected an exception is raised.
        """

        if not self._connected:
            raise Exception("Client not connected")

        self.lock.acquire()
        self._id += 1
        id_ = self._id
        self.lock.release()

        encoded_params = []
        for param in params:
            encoded_params.append(param)

        data = {
            "jsonrpc": "2.0",
            "id": self._id,
            "method": method,
            "handle": handle,
            "params": encoded_params,
        }
        data = reduce(lambda d, f: f(d), self._request_interceptors, data)
        json_data = json.dumps(data)
        self._socket.send(json_data)
        res = self._wait_response(id_)
        res = reduce(lambda r, f: f(r), self._response_interceptors, res)
        return_value = None
        if "result" in res:
            return_value = res["result"]
        elif "error" in res:
            raise Exception(res["error"]["message"])
        else:
            return_value = res
        return return_value

    def _wait_response(self, id_):
        """
        _wait_response waits (blocking) for a message with the specified id.
        Internal method that should only be called from send
        """

        with self._received_added:
            while id_ not in self._received:
                self._received_added.wait()
            res = self._received[id_]
            del self._received[id_]
            return res



from dataclasses import dataclass
import typing
import requests
from urllib.parse import urlparse

from .config import Config
from .rest import RestClient, ApiKeyAuth, Middleware
from .rpc import RpcClient


@dataclass
class Auth:
    """
    Auth can be used to make rest and rpc calls
    """

    config: Config
    rest: typing.Callable[
        [str, str, typing.Dict, typing.Dict, typing.Dict], requests.Response
    ]

    def __init__(self, config: Config, middlewares: typing.List[Middleware] = None):
        if middlewares is None:
            middlewares = []
        config.validate()
        self.config = config
        auth = ApiKeyAuth(api_key=self.config.api_key)
        rc = RestClient(base_url=self.config.host, auth=auth, middlewares=middlewares)
        self.rest = rc.rest

    def rpc(
        self,
        app_id: str,
        request_interceptors=None,
        response_interceptors=None,
    ) -> RpcClient:
        """
        rpc returns an RpcClient that can be used to
        connect to the engine for a specific app
        """
        hostname = urlparse(self.config.host).hostname
        ws_url = "wss://" + hostname.strip("/") + "/app/" + app_id
        header = ["Authorization: Bearer %s" % self.config.api_key]
        if request_interceptors is None:
            request_interceptors = []
        if response_interceptors is None:
            response_interceptors = []
        return RpcClient(
            ws_url,
            header,
            request_interceptors=request_interceptors,
            response_interceptors=response_interceptors,
        )



import requests
from dataclasses import dataclass, field
import typing


@dataclass
class Middleware:
    """
    Base class for creating request and response middleware to modify request and response
    """

    def handle_request(self, req: requests.Request):
        pass

    def handle_response(self, res: requests.Response):
        pass


@dataclass
class ApiKeyAuth(Middleware):
    """
    ApiKeyAuth is request middleware for authorizing using an api key
    """

    api_key: str

    def handle_request(self, req: requests.Request):
        req.headers["authorization"] = "Bearer " + self.api_key


class NoUrlException(Exception):
    """
    NoUrlException represents a missing URL
    """

    pass


class ConnectionException(Exception):
    """
    ConnectionException represents a failure in the request
    The request raised an exception
    ConnectionException is not a 4xx response
    """

    pass


class AuthenticationException(Exception):
    """
    AuthenticationException represents a failure to authenticate
    """

    pass


def _get_dict(o: any):
    """
    get_dict attempts to convert objects to dict, nested
    """
    try:
        d = o.__dict__
        for k in d:
            d[k] = _get_dict(d[k])
        return d
    except (Exception):
        return o


@dataclass
class RestClient:
    """
    RestClient represents a rest client for making api calls
    """

    base_url: str
    auth: Middleware = None
    middlewares: typing.List[Middleware] = field(default_factory=list)

    def rest(
        self,
        path: str,
        method="get",
        data=None,
        params: dict = None,
        headers: dict = None,
    ) -> requests.Response:
        """
        rest sends a request
        """
        if not self.base_url:
            raise NoUrlException("Caller has no 'base_url'")
        if not path.startswith("/api/v1/"):
            path = "/api/v1/" + path.strip("/")

        # If the data can be converted to a dict then send
        # it as json, otherwise send it as data.
        json_data = None
        if data:
            json_data = _get_dict(data)
            if json_data:
                data = None
        if params:
            params = _get_dict(params)

        # Create request.
        req = requests.Request(
            method,
            self.base_url + path,
            data=data,
            json=json_data,
            headers=headers,
            params=params,
        )

        # Use authentication middleware if provided.
        if self.auth:
            self.auth.handle_request(req)

        for mw in self.middlewares:
            mw.handle_request(req)

        with requests.Session() as session:
            prepared = req.prepare()
            try:
                res = session.send(prepared)
            except Exception as exc:
                raise ConnectionException("Connection Error: " + self.base_url) from exc
            for mw in self.middlewares:
                mw.handle_response(res)

            try:
                res.raise_for_status()
            except Exception as e:
                res.close()
                if res.status_code == 401:
                    raise AuthenticationException("Failed to authenticate")
                raise e

            return res
